//
// Copyright (c) Microsoft. All rights reserved.
// See https://aka.ms/azai/vision/license for the full license information.
//

#pragma once
#include <azac_api_c_properties.h>
#include <azac_api_cxx_common.h>
#include <azac_api_cxx_details_string_helpers.h>

// TODO: TFS#3671215 - Vision: C/C++ azac_api* files are in shared include directory, speech and vision share

/*! \cond INTERNAL */

namespace Azure {
namespace AI {
namespace Core {
namespace Details {

template <typename T>
struct PropertyCollectionKeyKind {
    static constexpr const char* GetKind()
    {
        static_assert(std::is_same<T, int>::value, "Didn't declare kind... See: PRIVATE_PROPERTY_COLLECTION_STATICS");
        return "int";
    }
};

/// <summary>
/// Represents a collection of properties and their string values.
/// </summary>
template <typename TPropKey = int>
class PropertyCollectionBase
{
public:

    /// <summary>
    /// Initializes a new instance of the PropertyCollectionBase class.
    /// </summary>
    static std::shared_ptr<AI::Core::Details::PropertyCollectionBase<TPropKey>> Create()
    {
        AZAC_HANDLE handle = AZAC_HANDLE_INVALID;
        AZAC_THROW_ON_FAIL(ai_core_properties_handle_create(&handle));
        return FromHandle(handle);
    }

    /// <summary>
    /// Destructs an instance of the PropertyCollectionBase class.
    /// </summary>
    ~PropertyCollectionBase()
    {
        if (ai_core_properties_handle_is_valid(m_properties))
        {
            ai_core_properties_handle_release(m_properties);
            m_properties = AZAC_HANDLE_INVALID;
        }
    }

    /// <summary>
    /// Sets the property string value for the specified key value.
    /// </summary>
    /// <param name="key">The string value's key</param>
    /// <param name="value">The new string value</param>
    void Set(TPropKey key, const char* value)
    {
        auto kind = Details::PropertyCollectionKeyKind<TPropKey>::GetKind();
        AZAC_THROW_ON_FAIL(ai_core_properties_string_set(m_properties, (int)key, kind, value));
    }

    /// <summary>
    /// Sets the property string value for the specified key value.
    /// </summary>
    /// <param name="key">The string value's key</param>
    /// <param name="value">The new string value</param>
    void Set(TPropKey key, const wchar_t* value)
    {
        Set(key, AI::Core::Details::to_string<std::string>(value));
    }

    /// <summary>
    /// Sets the property string value for the specified key value.
    /// </summary>
    /// <param name="key">The string value's key</param>
    /// <param name="value">The new string value</param>
    void Set(TPropKey key, const std::string& value) { Set(key, value.c_str()); }

    /// <summary>
    /// Sets the property string value for the specified key value.
    /// </summary>
    /// <param name="key">The string value's key</param>
    /// <param name="value">The new string value</param>
    void Set(TPropKey key, const std::wstring& value) { Set(key, AI::Core::Details::to_string<std::string>(value)); }

    /// <summary>
    /// Sets the property boolean value for the specified key value.
    /// </summary>
    /// <param name="key">The boolean value's key</param>
    /// <param name="value">The new boolean value</param>
    void Set(TPropKey key, bool value) { Set(key, std::to_string(value)); }

    /// <summary>
    /// Sets the property integer value for the specified key value.
    /// </summary>
    /// <param name="key">The integer value's key</param>
    /// <param name="value">The new integer value</param>
    void Set(TPropKey key, int value) { Set(key, std::to_string(value)); }

    /// <summary>
    /// Sets the property float value for the specified key value.
    /// </summary>
    /// <param name="key">The float value's key</param>
    /// <param name="value">The new float value</param>
    void Set(TPropKey key, float value) { Set(key, std::to_string(value)); }

    /// <summary>
    /// Sets the property string value for the specified key value.
    /// </summary>
    /// <param name="key">The string value's key</param>
    /// <param name="value">The new string value</param>
    void Set(const char* key, const char* value)
    {
        AZAC_THROW_ON_FAIL(ai_core_properties_string_set(m_properties, 0, key, value));
    }

    /// <summary>
    /// Sets the property string value for the specified key value.
    /// </summary>
    /// <param name="key">The string value's key</param>
    /// <param name="value">The new string value</param>
    void Set(const wchar_t* key, const wchar_t* value)
    {
        return Set(AI::Core::Details::to_string<std::string>(key), AI::Core::Details::to_string<std::string>(value));
    }

    /// <summary>
    /// Sets the property string value for the specified key value.
    /// </summary>
    /// <param name="key">The string value's key</param>
    /// <param name="value">The new string value</param>
    void Set(const std::string& key, const char* value) { Set(key.c_str(), value); }

    /// <summary>
    /// Sets the property string value for the specified key value.
    /// </summary>
    /// <param name="key">The string value's key</param>
    /// <param name="value">The new string value</param>
    void Set(const std::wstring& key, const wchar_t* value) { Set(key.c_str(), value); }

    /// <summary>
    /// Sets the property string value for the specified key value.
    /// </summary>
    /// <param name="key">The string value's key</param>
    /// <param name="value">The new string value</param>
    void Set(const std::string& key, const std::string& value) { Set(key.c_str(), value.c_str()); }

    /// <summary>
    /// Sets the property string value for the specified key value.
    /// </summary>
    /// <param name="key">The string value's key</param>
    /// <param name="value">The new string value</param>
    void Set(const std::wstring& key, const std::wstring& value) { Set(AI::Core::Details::to_string<std::string>(key), AI::Core::Details::to_string<std::string>(value)); }

    /// <summary>
    /// Sets the property boolean value for the specified key value.
    /// </summary>
    /// <param name="key">The boolean value's key</param>
    /// <param name="value">The new boolean value</param>
    void Set(const std::string& key, bool value) { Set(key, std::to_string(value)); }

    /// <summary>
    /// Sets the property boolean value for the specified key value.
    /// </summary>
    /// <param name="key">The boolean value's key</param>
    /// <param name="value">The new boolean value</param>
    void Set(const std::wstring& key, bool value) { Set(key, std::to_wstring(value)); }

    /// <summary>
    /// Sets the property integer value for the specified key value.
    /// </summary>
    /// <param name="key">The integer value's key</param>
    /// <param name="value">The new integer value</param>
    void Set(const std::string& key, int value) { Set(key, std::to_string(value)); }

    /// <summary>
    /// Sets the property integer value for the specified key value.
    /// </summary>
    /// <param name="key">The integer value's key</param>
    /// <param name="value">The new integer value</param>
    void Set(const std::wstring& key, int value) { Set(key, std::to_wstring(value)); }

    /// <summary>
    /// Sets the property integer value for the specified key value.
    /// </summary>
    /// <param name="key">The integer value's key</param>
    /// <param name="value">The new size_t value</param>
    void Set(const std::string& key, size_t value) { Set(key, std::to_string(value)); }

    /// <summary>
    /// Sets the property integer value for the specified key value.
    /// </summary>
    /// <param name="key">The integer value's key</param>
    /// <param name="value">The new size_t value</param>
    void Set(const std::wstring& key, size_t value) { Set(key, std::to_wstring(value)); }

    /// <summary>
    /// Gets the property string value for the specified key value.
    /// </summary>
    /// <param name="key">The string value's key value</param>
    /// <param name="defaultValue">A default string value, returned if the key value is not found</param>
    /// <returns>The property string value for the specified key value, or the default value if the key value is not found.</returns>
    const std::string Get(TPropKey key, const std::string& defaultValue = "") const
    {
        auto kind = Details::PropertyCollectionKeyKind<TPropKey>::GetKind();
        auto ptr = ai_core_properties_string_get(m_properties, (int)key, kind, defaultValue.c_str());
        auto value = std::string(ptr);
        AZAC_THROW_ON_FAIL(ai_core_properties_string_free(ptr));
        return value;
    }

    /// <summary>
    /// Gets the property string value for the specified key value.
    /// </summary>
    /// <param name="key">The string value's key value</param>
    /// <param name="defaultValue">A default string value, returned if the key value is not found</param>
    /// <returns>The property string value for the specified key value, or the default value if the key value is not found.</returns>
    const std::string Get(const std::string& key, const std::string& defaultValue = "") const
    {
        auto ptr = ai_core_properties_string_get(m_properties, 0, key.c_str(), defaultValue.c_str());
        auto value = std::string(ptr);
        AZAC_THROW_ON_FAIL(ai_core_properties_string_free(ptr));
        return value;
    }

    /// <summary>
    /// Gets the property binary value for the specified key value.
    /// </summary>
    /// <param name="key">The string value's key value</param>
    /// <param name="size">The size of the binary blob</param>
    /// <returns>The property binary value for the specified key value, or an empty shared_ptr if the key value is not found.</returns>
    std::shared_ptr<uint8_t> GetBinary(const std::string& key, uint32_t* size) const
    {
        auto ptr = ai_core_properties_binary_get(m_properties, 0, key.c_str(), size);
        if (ptr != nullptr) 
        {
            auto value = std::shared_ptr<uint8_t>(ptr, [](uint8_t* p) { AZAC_THROW_ON_FAIL(ai_core_properties_binary_free(p)); });
            return value;
        }

        return nullptr;
    }

protected:
    /*! \cond PROTECTED */
    static std::shared_ptr<AI::Core::Details::PropertyCollectionBase<TPropKey>> FromHandle(AZAC_HANDLE handle)
    {
        auto ptr = new AI::Core::Details::PropertyCollectionBase<TPropKey>(handle);
        return std::shared_ptr<AI::Core::Details::PropertyCollectionBase<TPropKey>>(ptr);
    }

    explicit PropertyCollectionBase(AZAC_HANDLE handle) :
        m_properties(handle)
    {
    }

    explicit operator AZAC_HANDLE() { return m_properties; }
    /*! \endcond */

private:

    AZAC_DISABLE_COPY_AND_MOVE(PropertyCollectionBase);

    AZAC_HANDLE m_properties;
};

template <typename TPropKey = int, typename TPropertyCollection = AI::Core::Details::PropertyCollectionBase<TPropKey>>
class PropertyCollection : public TPropertyCollection
{
public:

    PropertyCollection(AZAC_HANDLE handle, std::function<AZACHR(AZAC_HANDLE, AZAC_HANDLE*)> fn) :
        TPropertyCollection( [&]() {
            AZAC_HANDLE properties = AZAC_HANDLE_INVALID;
            AZAC_THROW_ON_FAIL(fn(handle, &properties));
            return properties;
        }())
    {
    }

    static std::shared_ptr<PropertyCollection<TPropKey, TPropertyCollection>> FromHandle(AZAC_HANDLE handle)
    {
        auto ptr = new PropertyCollection<TPropKey, TPropertyCollection>(handle, [](auto handle, auto* properties) { *properties = handle; return AZAC_ERR_NONE; });
        return std::shared_ptr<PropertyCollection<TPropKey, TPropertyCollection>>(ptr);
    }

    static std::shared_ptr<PropertyCollection<TPropKey, TPropertyCollection>> Create()
    {
        AZAC_HANDLE handle = AZAC_HANDLE_INVALID;
        AZAC_THROW_ON_FAIL(ai_core_properties_handle_create(&handle));
        return FromHandle(handle);
    }
};

} } } } // Azure::AI::Core::Details

#define PRIVATE_PROPERTY_COLLECTION_STATICS(x, name) \
namespace Azure { namespace AI { namespace Core { namespace Details { \
template<> \
constexpr const char* Azure::AI::Core::Details::PropertyCollectionKeyKind<x>::GetKind() { return name; } \
}}}}

/*! \endcond */
