//
// Copyright (c) Microsoft. All rights reserved.
// See https://aka.ms/azai/vision/license for the full license information.
//

#pragma once
#include <cstring>
#include <future>
#include <azac_api_c_async_op.h>
#include <azac_api_cxx_common.h>
#include <azac_api_cxx_details_promise_callback_helper.h>
#include <azac_api_cxx_details_property_collection.h>
#include <vision_api_cxx_common.h>
#include <vision_api_cxx_event_signal.h>
#include <vision_api_cxx_session.h>
#include <vision_api_cxx_source.h>
#include <vision_api_c_session.h>

namespace Azure {
namespace AI {
namespace Vision {
namespace Core {
namespace Details {

/// <summary>
/// Base class for all vision recognizers/analyzers
/// </summary>
template <typename TRecognizer,
          typename TOptions,
          typename TOptionKey,
          typename TStartOptions,
          typename TStartOptionKey,
          typename TProperties,
          typename TPropertyKey,
          typename TResult,
          typename TResultEventArgs,
          typename TRecognizingResult,
          typename TRecognizingResultEventArgs,
          typename TStartedEventArgs,
          typename TStoppedEventArgs>
class RecognizerBase
{
protected:
    /*! \cond PROTECTED */
    template<typename Target> using ProtectedAccess = Azure::AI::Core::Details::ProtectedAccess<Target>;
    AI::Core::Details::PropertyCollection<TPropertyKey, TProperties> m_properties;
    /*! \endcond */

public:

    template<class TConfigOptions>
    static std::shared_ptr<TRecognizer> Create(const char* viewKind, const std::shared_ptr<TConfigOptions>& serviceOptions,
        const std::shared_ptr<Input::VisionSource>& input, const std::shared_ptr<TOptions>& recognizerOptions = nullptr)
    {
        // create the session
        auto session = Core::Session::VisionSession::Create(serviceOptions, input);
        auto sessionHandle = ProtectedAccess<Core::Session::VisionSession>::HandleFromPtr(session.get());

        // create the view (aka Recognizer)
        AZAC_HANDLE view = AZAC_HANDLE_INVALID;
        AZAC_HANDLE optionsHandle = ProtectedAccess<TOptions>::HandleFromPtr(recognizerOptions.get());
        AZAC_THROW_ON_FAIL(vision_session_view_handle_create(&view, sessionHandle, viewKind, optionsHandle));
        auto ptr = TRecognizer::FromHandle(view);
        ptr->m_session = session;

        // hook up the event handlers
        AZAC_THROW_ON_FAIL(vision_event_callback_set(view, "session.started", ptr.get(), SessionEventCallbackHandler));
        AZAC_THROW_ON_FAIL(vision_event_callback_set(view, "session.stopped", ptr.get(), SessionEventCallbackHandler));
        AZAC_THROW_ON_FAIL(vision_event_callback_set(view, "recognizing", ptr.get(), SessionEventCallbackHandler));
        AZAC_THROW_ON_FAIL(vision_event_callback_set(view, "recognized", ptr.get(), SessionEventCallbackHandler));
        AZAC_THROW_ON_FAIL(vision_event_callback_set(view, "error", ptr.get(), SessionEventCallbackHandler));

        return ptr;
    }

    virtual ~RecognizerBase()
    {
        if (vision_session_view_handle_is_valid(m_recognizerView))
        {
            // release the event handlers
            AZAC_LOG_ON_FAIL(::vision_event_callback_set(m_recognizerView, "session.started", nullptr, nullptr));
            AZAC_LOG_ON_FAIL(::vision_event_callback_set(m_recognizerView, "session.stopped", nullptr, nullptr));
            AZAC_LOG_ON_FAIL(::vision_event_callback_set(m_recognizerView, "recognizing", nullptr, nullptr));
            AZAC_LOG_ON_FAIL(::vision_event_callback_set(m_recognizerView, "recognized", nullptr, nullptr));
            AZAC_LOG_ON_FAIL(::vision_event_callback_set(m_recognizerView, "error", nullptr, nullptr));

            ::vision_session_view_handle_release(m_recognizerView);
            m_recognizerView = AZAC_HANDLE_INVALID;
        }

        m_session.reset();
    };
    
    std::shared_ptr<Core::Session::VisionSession> GetSession() const
    {
        return m_session;
    }
    
    std::string GetSessionId() const
    {
        return m_session->GetSessionId();
    }

    template<class T = std::string>
    AI::Core::Details::enable_if_w_or_string_t<T> GetSessionId() const
    {
        return m_session->GetSessionId<T>();
    }

    std::shared_ptr<TResult> RecognizeOnce(const std::shared_ptr<TStartOptions>& options = nullptr)
    {
        auto timeout = GetMillisecondTimeout("session.recognize.once", 60000);

        AZAC_HANDLE startOptions = ProtectedAccess<TStartOptions>::HandleFromPtr(options.get());

        AZAC_HANDLE asyncOp = AZAC_HANDLE_INVALID;
        AZAC_THROW_ON_FAIL(vision_session_view_single_shot_start(m_recognizerView, startOptions, nullptr, &asyncOp));

        AZAC_HANDLE result = AZAC_HANDLE_INVALID;
        auto waited = async_op_wait_for_result(asyncOp, timeout, &result);
        AZAC_THROW_ON_FAIL(async_op_handle_release(asyncOp));

        auto recognized = VZ_SUCCEEDED(waited)
            ? ProtectedAccess<TResult>::FromHandle(result)
            : GetRecognizeOnceWaitFailedResult(waited);

        return recognized;
    }

    std::future<std::shared_ptr<TResult>> RecognizeOnceAsync(const std::shared_ptr<TStartOptions>& options = nullptr)
    {
        auto promiseResult = std::make_shared<std::promise<std::shared_ptr<TResult>>>();
        auto promiseAsyncOp = std::make_shared<std::promise<AZAC_HANDLE>>();

        auto ptr = static_cast<TRecognizer*>(this);
        auto asyncOpCallback = m_promises.PromiseAsyncOpCallback(ptr->shared_from_this(), nullptr, [=](int /* id */, AZAC_HANDLE /* asyncOpCallback */) {

            auto asyncOp = promiseAsyncOp->get_future().get();

            AZAC_HANDLE result = AZAC_HANDLE_INVALID;
            auto waited = async_op_wait_for_result(asyncOp, 0, &result);
            AZAC_REPORT_ON_FAIL(async_op_handle_release(asyncOp));

            auto recognized = VZ_SUCCEEDED(waited)
                ? ProtectedAccess<TResult>::FromHandle(result)
                : GetRecognizeOnceWaitFailedResult(waited);

            promiseResult->set_value(recognized);
        });

        AZAC_HANDLE startOptions = ProtectedAccess<TStartOptions>::HandleFromPtr(options.get());

        AZAC_HANDLE asyncOp = AZAC_HANDLE_INVALID;
        auto check = vision_session_view_single_shot_start(m_recognizerView, startOptions, asyncOpCallback, &asyncOp);

        if (AZAC_FAILED(check))
        {
            AZAC_REPORT_ON_FAIL(async_op_callback_handle_release(asyncOpCallback));
            AZAC_THROW_HR(check);
        }

        promiseAsyncOp->set_value(asyncOp);
        return promiseResult->get_future();
    }

    void StartContinuousRecognition(const std::shared_ptr<TStartOptions>& options = nullptr)
    {
        auto timeout = GetMillisecondTimeout("session.start.continuous.recognition", 60000);

        AZAC_HANDLE startOptions = ProtectedAccess<TStartOptions>::HandleFromPtr(options.get());

        AZAC_HANDLE asyncOp = AZAC_HANDLE_INVALID;
        AZAC_THROW_ON_FAIL(vision_session_view_continuous_start(m_recognizerView, startOptions, nullptr, &asyncOp));

        auto waited = async_op_wait_for(asyncOp, timeout);
        AZAC_THROW_ON_FAIL(async_op_handle_release(asyncOp));
        AZAC_THROW_ON_FAIL(waited);
    }

    std::future<void> StartContinuousRecognitionAsync(const std::shared_ptr<TStartOptions>& options = nullptr)
    {
        auto promise = std::make_shared<std::promise<void>>();

        auto ptr = static_cast<TRecognizer*>(this);
        auto asyncOpCallback = m_promises.PromiseAsyncOpCallback(ptr->shared_from_this(), std::move(promise));

        AZAC_HANDLE startOptions = ProtectedAccess<TStartOptions>::HandleFromPtr(options.get());

        AZAC_HANDLE asyncOp = AZAC_HANDLE_INVALID;
        auto check = vision_session_view_continuous_start(m_recognizerView, startOptions, asyncOpCallback, &asyncOp);
        AZAC_REPORT_ON_FAIL(async_op_handle_release(asyncOp));

        AZAC_IFTRUE(AZAC_FAILED(check), AZAC_REPORT_ON_FAIL(async_op_callback_handle_release(asyncOpCallback)));
        AZAC_THROW_ON_FAIL(check);

        return promise->get_future();
    }

    void StopContinuousRecognition()
    {
        auto timeout = GetMillisecondTimeout("session.stop.continuous.recognition", 60000);

        AZAC_HANDLE asyncOp = AZAC_HANDLE_INVALID;
        AZAC_THROW_ON_FAIL(vision_session_view_continuous_stop(m_recognizerView, nullptr, nullptr, &asyncOp));

        auto waited = async_op_wait_for(asyncOp, timeout);
        AZAC_THROW_ON_FAIL(async_op_handle_release(asyncOp));
        AZAC_THROW_ON_FAIL(waited);
    }

    std::future<void> StopContinuousRecognitionAsync()
    {
        auto promise = std::make_shared<std::promise<void>>();

        auto ptr = static_cast<TRecognizer*>(this);
        auto asyncOpCallback = m_promises.PromiseAsyncOpCallback(ptr->shared_from_this(), std::move(promise));

        AZAC_HANDLE asyncOp = AZAC_HANDLE_INVALID;
        auto check = vision_session_view_continuous_stop(m_recognizerView, nullptr, asyncOpCallback, &asyncOp);
        AZAC_REPORT_ON_FAIL(async_op_handle_release(asyncOp));

        AZAC_IFTRUE(AZAC_FAILED(check), AZAC_REPORT_ON_FAIL(async_op_callback_handle_release(asyncOpCallback)));
        AZAC_THROW_ON_FAIL(check);

        return promise->get_future();
    }

    void StopRecognitionOnce()
    {
        auto timeout = GetMillisecondTimeout("session.stop.once.recognition", 60000);
        AZAC_HANDLE asyncOp = AZAC_HANDLE_INVALID;
        AZAC_THROW_ON_FAIL(vision_session_view_single_shot_stop(m_recognizerView, nullptr, nullptr, &asyncOp));

        auto waited = async_op_wait_for(asyncOp, timeout);
        AZAC_THROW_ON_FAIL(async_op_handle_release(asyncOp));
        AZAC_THROW_ON_FAIL(waited);
    }

    std::future<void> StopRecognitionOnceAsync()
    {
        auto promise = std::make_shared<std::promise<void>>();

        auto ptr = static_cast<TRecognizer*>(this);
        auto asyncOpCallback = m_promises.PromiseAsyncOpCallback(ptr->shared_from_this(), std::move(promise));

        AZAC_HANDLE asyncOp = AZAC_HANDLE_INVALID;
        auto check = vision_session_view_single_shot_stop(m_recognizerView, nullptr, asyncOpCallback, &asyncOp);
        AZAC_REPORT_ON_FAIL(async_op_handle_release(asyncOp));

        AZAC_IFTRUE(AZAC_FAILED(check), AZAC_REPORT_ON_FAIL(async_op_callback_handle_release(asyncOpCallback)));
        AZAC_THROW_ON_FAIL(check);

        return promise->get_future();
    }

    bool WaitForStop()
    {
        auto timeout = std::chrono::milliseconds(GetMillisecondTimeout("session.wait.for.stop", 60000));
        return WaitForStop(timeout);
    }

    bool WaitForStop(const std::chrono::milliseconds& timeout)
    {
        auto future = WaitForStopAsync();
        auto status = future.wait_for(timeout);
        return status == std::future_status::ready;
    }

    std::future<void> WaitForStopAsync()
    {
        auto stoppedPromise = std::make_shared<std::promise<void>>();
        if (m_started.load())
        {
            auto ptr = static_cast<TRecognizer*>(this);
            m_promises.PromiseCallback(ptr->shared_from_this(), "session.stopped", std::move(stoppedPromise), [=](int id, AZAC_HANDLE /* handle */) {
                m_promises.RemovePromiseCallback(id, "session.stopped");
            });
        }
        else
        {
            stoppedPromise->set_value();
        }

        return stoppedPromise->get_future();
    }

protected:

    /*! \cond PROTECTED */
    Core::EventSignal<const TStartedEventArgs&> m_SessionStarted;
    Core::EventSignal<const TStoppedEventArgs&> m_SessionStopped;
    Core::EventSignal<const TRecognizingResultEventArgs&> m_Recognizing;
    Core::EventSignal<const TResultEventArgs&> m_Recognized;

    TProperties& GetProperties() { return m_properties; }
    const TProperties& GetProperties() const { return m_properties; }
    /*! \endcond */

public:

    static std::shared_ptr<TRecognizer> FromHandle(AZAC_HANDLE handle)
    {
        auto ptr = new TRecognizer(handle);
        return std::shared_ptr<TRecognizer>(ptr);
    }

    explicit RecognizerBase(AZAC_HANDLE view) :
        m_properties(view, [](auto handle, auto* properties) { return vision_session_view_properties_handle_get(handle, properties); }),
        m_recognizerView(view)
    {
    }

    explicit operator AZAC_HANDLE() { return m_recognizerView; }

protected:

    /*! \cond PROTECTED */
    uint32_t GetMillisecondTimeout(const char* prefix, int defaultMilliseconds)
    {
        auto name = std::string(prefix) + ".timeout.milliseconds";
        auto value = m_properties.Get(name, std::to_string(defaultMilliseconds));
        return std::stoi(value);
    }

    std::shared_ptr<TResult> GetRecognizeOnceWaitFailedResult(AZACHR waited)
    {
        auto timeout = GetMillisecondTimeout("session.recognize.once.wait.failed.stop.timeout", 2000);

        AZAC_HANDLE asyncOp = AZAC_HANDLE_INVALID;
        AZAC_REPORT_ON_FAIL(vision_session_view_single_shot_stop(m_recognizerView, nullptr, nullptr, &asyncOp));
        AZAC_REPORT_ON_FAIL(async_op_wait_for(asyncOp, timeout));
        AZAC_REPORT_ON_FAIL(async_op_handle_release(asyncOp));

        return GetWaitFailedResult(waited);
    }

    std::shared_ptr<TResult> GetWaitFailedResult(AZACHR waited)
    {
        return waited == AZAC_ERR_TIMEOUT
            ? GetWaitTimedoutResult()
            : GetWaitErrorResult(waited);
    }

    std::shared_ptr<TResult> GetWaitTimedoutResult()
    {
        // TODO: TFS#3664100 - Vision: RecognizeOnce timeout/failed results implementation and tests
        return std::shared_ptr<TResult>(nullptr);
    }

    std::shared_ptr<TResult> GetWaitErrorResult(AZACHR waited)
    {
        // TODO: TFS#3664100 - Vision: RecognizeOnce timeout/failed results implementation and tests
        if (VZ_FAILED(waited))
        {
            // This code will be removed soon after some refactoring of the input media file source.
            // The desired behavior is to return a TResult object with appropriate error instead of throwing.
            AZAC_HANDLE handle = reinterpret_cast<AZAC_HANDLE>(waited);
            const std::string callstack(error_get_call_stack(handle));
            const std::string message(error_get_message(handle));
            std::runtime_error e(message + callstack);
            throw e;
        }

        return nullptr;
    }

    static void SessionEventCallbackHandler(AZAC_HANDLE session, const char* name, void* context, AZAC_HANDLE eventArgs)
    {
        auto ptr = static_cast<TRecognizer*>(context);
        ptr->m_promises.CompletePromiseCallbacks(name, eventArgs);

        if (std::strcmp(name, "session.started") == 0)
            ptr->SessionStartedCallback(session, context, eventArgs);
        else if (std::strcmp(name, "session.stopped") == 0)
            ptr->SessionStoppedCallback(session, context, eventArgs);
        else if (std::strcmp(name, "recognizing") == 0)
            ptr->RecognizingCallback(session, context, eventArgs);
        else if(std::strcmp(name, "recognized") == 0)
            ptr->RecognizedCallback(session, context, eventArgs);
        else if(std::strcmp(name, "error") == 0)
            ptr->RecognizedCallback(session, context, eventArgs);
    }

    void SessionStartedCallback(AZAC_HANDLE session, void* context, AZAC_HANDLE eventArgs)
    {
        UNUSED(session); UNUSED(context);

        auto startedBefore = m_started.exchange(true);
        AZAC_DBG_ASSERT(!startedBefore); UNUSED(startedBefore);

        auto connected = m_SessionStarted.IsConnected();
        if (!connected)
        {
            return;
        }

        auto ptr = ProtectedAccess<TStartedEventArgs>::FromHandle(eventArgs);
        m_SessionStarted.Signal(*ptr.get());
    }

    void SessionStoppedCallback(AZAC_HANDLE session, void* context, AZAC_HANDLE eventArgs)
    {
        UNUSED(session); UNUSED(context);

        auto startedBefore = m_started.exchange(false);
        AZAC_DBG_ASSERT(startedBefore); UNUSED(startedBefore);

        auto connected = m_SessionStopped.IsConnected();
        if (!connected)
        {
            return;
        }

        auto ptr = ProtectedAccess<TStoppedEventArgs>::FromHandle(eventArgs);
        m_SessionStopped.Signal(*ptr.get());
    }

    void RecognizingCallback(AZAC_HANDLE session, void* context, AZAC_HANDLE eventArgs)
    {
        UNUSED(session); UNUSED(context);

        auto connected = m_Recognizing.IsConnected();
        if (!connected)
        {
            return;
        }

        auto ptr = ProtectedAccess<TRecognizingResultEventArgs>::FromHandle(eventArgs);
        m_Recognizing.Signal(*ptr.get());
    }

    void RecognizedCallback(AZAC_HANDLE session, void* context, AZAC_HANDLE eventArgs)
    {
        UNUSED(session); UNUSED(context);

        auto connected = m_Recognized.IsConnected();
        if (!connected)
        {
            return;
        }

        auto ptr = ProtectedAccess<TResultEventArgs>::FromHandle(eventArgs);
        m_Recognized.Signal(*ptr.get());
    }
    /*! \endcond */

private:

    AZAC_DISABLE_DEFAULT_CTORS(RecognizerBase);

    AZAC_HANDLE m_recognizerView;
    std::shared_ptr<Core::Session::VisionSession> m_session;

    std::atomic_bool m_started { false };
    AI::Core::Details::PromiseCallbackHelper<TRecognizer> m_promises;
};

} } } } } // Azure::AI::Vision::Core::Details
