//
// Copyright (c) Microsoft. All rights reserved.
// See https://aka.ms/azai/vision/license for the full license information.
//

#pragma once
#include <azac_api_cxx_details_property_collection.h>
#include <vision_api_cxx_common.h>
#include <vision_api_cxx_service_advanced_options.h>

#include <vector>
using namespace Azure::AI::Vision;

namespace Azure {
namespace AI {
namespace Vision {
namespace Embedded {

/// <summary>
/// Represents the Model related configuration property
/// </summary>
/// <remarks>
/// </remarks>
class VisionModel
{
public:
    /// <summary>
    /// Initializes a new instance of the VisionModel class with specified offline model configuration (path to config file and decryption key)
    /// </summary>
    /// <param name="path">The path to model configuration file</param>
    /// <param name="key">decryption key.</param>
    /// <returns>The newly created VisionModel</returns>
    VisionModel(const std::string& path, const std::string& key) :
        m_path(path), m_key(key) {}

    /// <summary>
    /// Initializes a new instance of the VisionModel class with specified offline model configuration (path to config file and decryption key)
    /// </summary>
    /// <param name="path">The path to model configuration file</param>
    /// <returns>The newly created VisionModel</returns>
    VisionModel(const std::string& path) :
        m_path(path), m_key("") {}

    /// <summary>
    /// Initializes a new instance of the VisionModel class with specified offline model configuration (path to config file and decryption key)
    /// </summary>
    /// <param name="path">The path to model configuration file</param>
    /// <param name="key">decryption key.</param>
    /// <returns>The newly created VisionModel</returns>
    VisionModel(const std::wstring& path, const std::wstring& key) :
        m_path(AI::Core::Details::to_string<std::string>(path)), m_key(AI::Core::Details::to_string<std::string>(key)) {}

    /// <summary>
    /// Initializes a new instance of the VisionModel class with specified offline model configuration (path to config file and decryption key)
    /// </summary>
    /// <param name="path">The path to model configuration file</param>
    /// <returns>The newly created VisionModel</returns>
    VisionModel(const std::wstring& path) :
        m_path(AI::Core::Details::to_string<std::string>(path)), m_key("") {}

    /// <summary>
    /// Get the path to model configuration file
    /// </summary>
    /// <returns>The path to model configuration file</returns>
    std::string GetPath() const
    {
        return m_path;
    }

    /// <summary>
    /// Get the model decryption key
    /// </summary>
    /// <returns>model decryption key</returns>
    std::string GetKey() const
    {
        return m_key;
    }

private:
    std::string m_path;
    std::string m_key;
};

/// <summary>
/// Represents the configuration options used to load embedded vision models
/// </summary>
/// <remarks>
/// Use one of the static methods VisionEmbeddedOptions::FromModel()
/// to instantiate an object of this class
/// </remarks>
class EmbeddedVisionOptions
{
private:
    template<typename Target> using ProtectedAccess = ::Azure::AI::Core::Details::ProtectedAccess<Target>;
    std::shared_ptr<Service::VisionServiceAdvancedOptions> m_options;
public:

    /// <summary>
    /// Initializes a new instance of the EmbeddedVisionOptions class with specified offline model config (path to config file and decryption key)
    /// </summary>
    /// <param name="modelPath">The configuration object with path to model configuration file and optional decryption key.</param>
    /// <returns>The newly created EmbeddedVisionOptions wrapped inside a std::shared_ptr</returns>
    static std::shared_ptr<EmbeddedVisionOptions> FromModel(const VisionModel& modelPath)
    {
        auto ptr = new EmbeddedVisionOptions();
        ptr->Advanced.Set("reco.engine.adapter", "embedded-ocr");
        ptr->Advanced.Set("embedded.vision.modelconfigurationpath", modelPath.GetPath().c_str());
        ptr->Advanced.Set("embedded.vision.modeldecryptionkey", modelPath.GetKey().c_str());

        return std::shared_ptr<EmbeddedVisionOptions>(ptr);
    }

    /// <summary>
    /// Destructs an instance of the EmbeddedVisionOptions class.
    /// </summary>
    virtual ~EmbeddedVisionOptions() = default;

    /// <summary>
    /// Advanced options and parameters
    /// </summary>
    Service::VisionServiceAdvancedOptions& Advanced;
protected:
    explicit EmbeddedVisionOptions(): m_options(ProtectedAccess<Service::VisionServiceAdvancedOptions>::Create()),
        Advanced(*m_options.get())
    {
    }
    explicit operator AZAC_HANDLE()
    {
        return ProtectedAccess<Service::VisionServiceAdvancedOptions>::HandleFromPtr(m_options.get());
    }

private:

    AZAC_DISABLE_COPY_AND_MOVE(EmbeddedVisionOptions);
};

} } } } // Azure::AI::Vision::Embedded
