//
// Copyright (c) Microsoft. All rights reserved.
// See https://aka.ms/azai/vision/license for the full license information.
//

#pragma once

#include <azac_api_cxx_details_lifetime.h>
#include <azac_api_cxx_details_property_collection.h>
#include <vision_api_cxx_common.h>
#include <vision_api_cxx_frame.h>
#include <vision_api_c_frame_writer.h>

namespace Azure {
namespace AI {
namespace Vision {
namespace Input {
namespace Frames {

// TODO: TFS#3664413 - What to do about stream parameter? (feedback from ryan)

/// <summary>
/// Represents the ability to write image frame data, for use as input with Vision AI scenario operations.
/// </summary>
class FrameWriter
{
public:

    /// <summary>
    /// Destructs an instance of the FrameWriter class.
    /// </summary>
    virtual ~FrameWriter()
    {
        if (vision_frame_writer_handle_is_valid(m_frameWriter))
        {
            vision_frame_writer_handle_release(m_frameWriter);
            m_frameWriter = AZAC_HANDLE_INVALID;
        }
    }

    /// <summary>
    /// Writes a single frame of image data to the underlying FrameSource.
    /// </summary>
    /// <param name="data">The data buffer.</param>
    /// <param name="dataSizeInBytes">The data buffer size in bytes</param>
    void WriteFrame(const uint8_t* data, size_t dataSizeInBytes)
    {
        uint32_t size = (uint32_t)dataSizeInBytes; // idiomatic to accept size_t, but C API uses uint32_t
        AZAC_THROW_ON_FAIL(vision_frame_writer_write(m_frameWriter, 0, data, size, nullptr, nullptr));
    }

    /// <summary>
    /// Writes a single frame of image data to the underlying FrameSource.
    /// </summary>
    /// <param name="frame">A single image frame.</param>
    void WriteFrame(std::shared_ptr<Frame> frame)
    {
        uint32_t size = (uint32_t)frame->DataSizeInBytes; // idiomatic to accept size_t, but C API uses uint32_t
        AZAC_THROW_ON_FAIL(vision_frame_writer_write_with_properties(m_frameWriter, 0, frame->Data, size, AI::Core::Details::ProtectedAccess<Frame>::HandleFromPtr(frame.get())));
    }

protected:
    /*! \cond PROTECTED */
    static std::shared_ptr<FrameWriter> FromHandle(AZAC_HANDLE handle, const Azure::AI::Core::Details::LifetimeBinder& binder)
    {
        auto ptr = new FrameWriter(handle, binder);
        return std::shared_ptr<FrameWriter>(ptr);
    }

    explicit FrameWriter(AZAC_HANDLE frameWriter, const Azure::AI::Core::Details::LifetimeBinder& binder) :
        m_frameWriter(frameWriter),
        m_binder{ binder }
    {
    }

    explicit operator AZAC_HANDLE() { return m_frameWriter; }
    /*! \endcond */

private:

    AZAC_DISABLE_DEFAULT_CTORS(FrameWriter);

    AZAC_HANDLE m_frameWriter;
    Azure::AI::Core::Details::LifetimeBinder m_binder;
};

} } } } } // Azure::AI::Vision::Input::Frames
