//
// Copyright (c) Microsoft. All rights reserved.
// See https://aka.ms/azai/vision/license for the full license information.
//

#pragma once
#include <azac_api_cxx_details_property_collection.h>
#include <vision_api_cxx_common.h>
#include <vision_api_cxx_service_option.h>

namespace Azure {
namespace AI {
namespace Vision {
namespace Service {

/// <summary>
/// Represents advanced options and parameters on the VisionService object.
/// </summary>
class VisionServiceAdvancedOptions
{
private:

    template<typename Target> using ProtectedAccess = Azure::AI::Core::Details::ProtectedAccess<Target>;
    using PropertyCollection = AI::Core::Details::PropertyCollection<VisionServiceOption>;
    std::shared_ptr<PropertyCollection> m_options;

public:

    /// <summary>
    /// Sets the string value for the specified option.
    /// </summary>
    /// <param name="option">The option key</param>
    /// <param name="value">The new string value</param>
    void Set(VisionServiceOption option, const char* value)
    {
        m_options->Set(option, value);
    }

    /// <summary>
    /// Sets the string value for the specified option.
    /// </summary>
    /// <param name="option">The option key</param>
    /// <param name="value">The new string value</param>
    void Set(VisionServiceOption option, const wchar_t* value)
    {
        m_options->Set(option, value);
    }

    /// <summary>
    /// Sets the string value for the specified option.
    /// </summary>
    /// <param name="option">The option key</param>
    /// <param name="value">The new string value</param>
    void Set(const std::string& option, const char* value)
    {
        m_options->Set(option, value);
    }

    /// <summary>
    /// Sets the string value for the specified option.
    /// </summary>
    /// <param name="option">The option key</param>
    /// <param name="value">The new string value</param>
    void Set(const std::wstring& option, const wchar_t* value)
    {
        m_options->Set(option, value);
    }

    /// <summary>
    /// Sets the string value for the specified option.
    /// </summary>
    /// <param name="option">The option key</param>
    /// <param name="value">The new string value</param>
    void Set(VisionServiceOption option, const std::string& value)
    {
        m_options->Set(option, value);
    }

    /// <summary>
    /// Sets the string value for the specified option.
    /// </summary>
    /// <param name="option">The option key</param>
    /// <param name="value">The new string value</param>
    void Set(VisionServiceOption option, const std::wstring& value)
    {
        m_options->Set(option, value);
    }

    /// <summary>
    /// Sets the property boolean value for the specified key value.
    /// </summary>
    /// <param name="option">The option key</param>
    /// <param name="value">The new boolean value</param>
    void Set(VisionServiceOption option, bool value)
    {
        m_options->Set(option, value);
    }

    /// <summary>
    /// Sets the property integer value for the specified key value.
    /// </summary>
    /// <param name="option">The option key</param>
    /// <param name="value">The new integer value</param>
    void Set(VisionServiceOption option, int value)
    {
        m_options->Set(option, value);
    }

    /// <summary>
    /// Sets the string value for the specified option.
    /// </summary>
    /// <param name="option">The option key</param>
    /// <param name="value">The new string value</param>
    void Set(const std::string& option, const std::string& value)
    {
        m_options->Set(option, value);
    }

    /// <summary>
    /// Sets the string value for the specified option.
    /// </summary>
    /// <param name="option">The option key</param>
    /// <param name="value">The new string value</param>
    void Set(const std::wstring& option, const std::wstring& value)
    {
        m_options->Set(option, value);
    }

    /// <summary>
    /// Sets the property boolean value for the specified key value.
    /// </summary>
    /// <param name="option">The option key</param>
    /// <param name="value">The new boolean value</param>
    void Set(const std::string& option, bool value)
    {
        m_options->Set(option, value);
    }

    /// <summary>
    /// Sets the property boolean value for the specified key value.
    /// </summary>
    /// <param name="option">The option key</param>
    /// <param name="value">The new boolean value</param>
    void Set(const std::wstring& option, bool value)
    {
        m_options->Set(option, value);
    }

    /// <summary>
    /// Sets the property integer value for the specified key value.
    /// </summary>
    /// <param name="option">The option key</param>
    /// <param name="value">The new integer value</param>
    void Set(const std::string& option, int value)
    {
        m_options->Set(option, value);
    }

    /// <summary>
    /// Sets the property integer value for the specified key value.
    /// </summary>
    /// <param name="option">The option key</param>
    /// <param name="value">The new integer value</param>
    void Set(const std::wstring& option, int value)
    {
        m_options->Set(option, value);
    }

    /// <summary>
    /// Gets an advanced option by its name
    /// </summary>
    /// <param name="option">The name of the advanced option to get</param>
    /// <param name="defaultValue">The value to be returned if the option was not set</param>
    /// <returns>The value of the option, or the specified default value if this option was not set</returns>
    const std::string Get(const std::string& option, const std::string& defaultValue = "") const
    {
        return m_options->Get(option, defaultValue);
    }

protected:
    /*! \cond PROTECTED */
    static std::shared_ptr<VisionServiceAdvancedOptions> Create()
    {
        auto ptr = new VisionServiceAdvancedOptions();
        return std::shared_ptr<VisionServiceAdvancedOptions>(ptr);
    }

    explicit VisionServiceAdvancedOptions() :
        m_options(PropertyCollection::Create())
    {
    }

    explicit operator AZAC_HANDLE()
    {
        return ProtectedAccess<PropertyCollection>::HandleFromPtr(m_options.get());
    }
    /*! \endcond */

private:

    AZAC_DISABLE_COPY_AND_MOVE(VisionServiceAdvancedOptions);
};

} } } } // Azure::AI::Vision::Service
