//
// Copyright (c) Microsoft. All rights reserved.
// See https://aka.ms/azai/vision/license for the full license information.
//

#pragma once
#include <azac_api_cxx_details_property_collection.h>
#include <vision_api_cxx_common.h>
#include <vision_api_cxx_service_advanced_options.h>
#include <vision_api_cxx_service_option.h>

namespace Azure {
namespace AI {
namespace Vision {
namespace Service {

/// <summary>
/// Represents the configuration options used to connect to a remote Vision AI Service
/// </summary>
/// <remarks>
/// Use one of the static methods VisionServiceOptions::FromEndpoint()
/// to instantiate an object of this class
/// </remarks>
class VisionServiceOptions
{
private:

    template<typename Target> using ProtectedAccess = ::Azure::AI::Core::Details::ProtectedAccess<Target>;
    std::shared_ptr<VisionServiceAdvancedOptions> m_options;

public:

    /// <summary>
    /// Initializes a new instance of the VisionServiceOptions class used to connect to the specified URL endpoint.
    /// </summary>
    /// <param name="endpoint">The endpoint URL of the Vision Service</param>
    /// <returns>The newly created VisionServiceOptions wrapped inside a std::shared_ptr</returns>
    static std::shared_ptr<VisionServiceOptions> FromEndpoint(const std::string& endpoint)
    {
        auto ptr = new VisionServiceOptions();
        ptr->Advanced.Set("service.endpoint", endpoint.c_str());

        return std::shared_ptr<VisionServiceOptions>(ptr);
    }

    /// <summary>
    /// Initializes a new instance of the VisionServiceOptions class used to connect to the specified URL endpoint.
    /// </summary>
    /// <param name="endpoint">The endpoint URL of the Vision Service</param>
    /// <param name="key">The Computer Vision key used to authenticate against the Vision Service</param>
    /// <returns>The newly created VisionServiceOptions wrapped inside a std::shared_ptr</returns>
    /// <remarks>
    /// Store your key securely. For example, using Azure Key Vault. Do not compile the value of your key into
    /// your application source code. Fetch it at run-time when needed from a secure location.
    /// </remarks>
    static std::shared_ptr<VisionServiceOptions> FromEndpoint(const std::string& endpoint, const std::string& key)
    {
        auto ptr = new VisionServiceOptions();
        ptr->Advanced.Set("service.endpoint", endpoint.c_str());
        ptr->Advanced.Set("service.auth.key", key.c_str());

        return std::shared_ptr<VisionServiceOptions>(ptr);
    }

    /// <summary>
    /// Initializes a new instance of the VisionServiceOptions class used to connect to the specified URL endpoint.
    /// </summary>
    /// <param name="endpoint">The endpoint URL of the Vision Service</param>
    /// <returns>The newly created VisionServiceOptions wrapped inside a std::shared_ptr</returns>
    static std::shared_ptr<VisionServiceOptions> FromEndpoint(const std::wstring& endpoint)
    {
        return FromEndpoint(AI::Core::Details::to_string<std::string>(endpoint));
    }

    /// <summary>
    /// Initializes a new instance of the VisionServiceOptions class used to connect to the specified URL endpoint.
    /// </summary>
    /// <param name="endpoint">The endpoint URL of the Vision Service</param>
    /// <param name="key">The Computer Vision key used to authenticate against the Vision Service</param>
    /// <returns>The newly created VisionServiceOptions wrapped inside a std::shared_ptr</returns>
    /// <remarks>
    /// Store your key securely. For example, using Azure Key Vault. Do not compile the value of your key into
    /// your application source code. Fetch it at run-time when needed from a secure location.
    /// </remarks>
    static std::shared_ptr<VisionServiceOptions> FromEndpoint(const std::wstring& endpoint, const std::wstring& key)
    {
        return FromEndpoint(
            AI::Core::Details::to_string<std::string>(endpoint),
            AI::Core::Details::to_string<std::string>(key));
    }

    /// <summary>
    /// Sets a new authorization token to be used to connect to the Vision Service.
    /// </summary>
    /// <param name="token">The authorization token</param>
    /// <remarks>
    /// If the previously given token has expired, you must call this method with a new token before
    /// a new connections to the Vision Service can be made.
    /// </remarks>
    void SetAuthorizationToken(const std::string& token)
    {
        Advanced.Set("service.auth.token", token);
    }

    /// <summary>
    /// Sets a new authorization token to be used to connect to the Vision Service.
    /// </summary>
    /// <param name="token">The authorization token</param>
    /// <remarks>
    /// If the previously given token has expired, you must call this method with a new token before
    /// a new connections to the Vision Service can be made.
    /// </remarks>
    void SetAuthorizationToken(const std::wstring& token)
    {
        SetAuthorizationToken(AI::Core::Details::to_string<std::string>(token));
    }

    /// <summary>
    /// Destructs an instance of the VisionServiceOptions class.
    /// </summary>
    virtual ~VisionServiceOptions() = default;

    /// <summary>
    /// Advanced options and parameters
    /// </summary>
    VisionServiceAdvancedOptions& Advanced;

    // TODO: TFS#3664466 - Vision: What VisionServiceOptions strongly typed methods go here?

protected:
    /*! \cond PROTECTED */
    explicit VisionServiceOptions() :
        m_options(ProtectedAccess<VisionServiceAdvancedOptions>::Create()),
        Advanced(*m_options.get())
    {
    }

    explicit operator AZAC_HANDLE()
    {
        return ProtectedAccess<VisionServiceAdvancedOptions>::HandleFromPtr(m_options.get());
    }
    /*! \endcond */

private:

    AZAC_DISABLE_COPY_AND_MOVE(VisionServiceOptions);
};

} } } } // Azure::AI::Vision::Service
