//
// Copyright (c) Microsoft. All rights reserved.
// See https://aka.ms/azai/vision/license for the full license information.
//

#pragma once
#include <azac_api_cxx_details_property_collection.h>
#include <vision_api_cxx_common.h>
#include <vision_api_cxx_session_option.h>
#include <vision_api_cxx_session_result_reason.h>
#include <vision_api_c_result.h>

namespace Azure {
namespace AI {
namespace Vision {
namespace Core {
namespace Session {

/// <summary>
/// Represents the options and parameters used to initialize a VisionSession instance.
/// </summary>
class VisionSessionOptions
{
public:

    void SetDeploymentId(const char* deploymentId)
    {
        m_options->Set(VisionSessionOption::DeploymentId, deploymentId);
    }

    void SetOperationId(const char* operationId)
    {
        m_options->Set(VisionSessionOption::OperationId, operationId);
    }

    void SetOperationConfigurationJson(const char* operationConfigurationJson)
    {
        m_options->Set(VisionSessionOption::OperationConfigurationJson, operationConfigurationJson);
    }

    void SetOperationConfigurationFilePath(const char* operationConfigurationFilePath)
    {
        m_options->Set(VisionSessionOption::OperationConfigurationFilePath, operationConfigurationFilePath);
    }

    /// <summary>
    /// Represents advanced options and parameters used to initialize a VisionSession instance.
    /// </summary>
    class AdvancedOptions
    {
    private:

        template<typename Target> using ProtectedAccess = Azure::AI::Core::Details::ProtectedAccess<Target>;
        using PropertyCollection = AI::Core::Details::PropertyCollection<VisionSessionOption>;
        std::shared_ptr<PropertyCollection> m_options;

    public:

        /// <summary>
        /// Sets the string value for the specified option.
        /// </summary>
        /// <param name="option">The vision session option</param>
        /// <param name="value">The new string value</param>
        void Set(VisionSessionOption option, const char* value)
        {
            m_options->Set(option, value);
        }

        /// <summary>
        /// Sets the string value for the specified option.
        /// </summary>
        /// <param name="option">The vision session option</param>
        /// <param name="value">The new string value</param>
        void Set(VisionSessionOption option, const wchar_t* value)
        {
            m_options->Set(option, value);
        }

        /// <summary>
        /// Sets the string value for the specified option.
        /// </summary>
        /// <param name="option">The vision session option</param>
        /// <param name="value">The new string value</param>
        void Set(VisionSessionOption option, const std::string& value)
        {
            m_options->Set(option, value);
        }

        /// <summary>
        /// Sets the string value for the specified option.
        /// </summary>
        /// <param name="option">The vision session option</param>
        /// <param name="value">The new string value</param>
        void Set(VisionSessionOption option, const std::wstring& value)
        {
            m_options->Set(option, value);
        }

        /// <summary>
        /// Sets the property boolean value for the specified key value.
        /// </summary>
        /// <param name="option">The vision session option</param>
        /// <param name="value">The new boolean value</param>
        void Set(VisionSessionOption option, bool value)
        {
            m_options->Set(option, value);
        }

        /// <summary>
        /// Sets the property integer value for the specified key value.
        /// </summary>
        /// <param name="option">The vision session option</param>
        /// <param name="value">The new integer value</param>
        void Set(VisionSessionOption option, int value)
        {
            m_options->Set(option, value);
        }

        /// <summary>
        /// Sets the string value for the specified option.
        /// </summary>
        /// <param name="option">The string option</param>
        /// <param name="value">The new string value</param>
        void Set(const std::string& option, const char* value)
        {
            m_options->Set(option, value);
        }

        /// <summary>
        /// Sets the string value for the specified option.
        /// </summary>
        /// <param name="option">The string option</param>
        /// <param name="value">The new string value</param>
        void Set(const std::wstring& option, const wchar_t* value)
        {
            m_options->Set(option, value);
        }

        /// <summary>
        /// Sets the string value for the specified option.
        /// </summary>
        /// <param name="option">The string option</param>
        /// <param name="value">The new string value</param>
        void Set(const std::string& option, const std::string& value)
        {
            m_options->Set(option, value);
        }

        /// <summary>
        /// Sets the string value for the specified option.
        /// </summary>
        /// <param name="option">The string option</param>
        /// <param name="value">The new string value</param>
        void Set(const std::wstring& option, const std::wstring& value)
        {
            m_options->Set(option, value);
        }

        /// <summary>
        /// Sets the property boolean value for the specified key value.
        /// </summary>
        /// <param name="option">The string option</param>
        /// <param name="value">The new boolean value</param>
        void Set(const std::string& option, bool value)
        {
            m_options->Set(option, value);
        }

        /// <summary>
        /// Sets the property boolean value for the specified key value.
        /// </summary>
        /// <param name="option">The string option</param>
        /// <param name="value">The new boolean value</param>
        void Set(const std::wstring& option, bool value)
        {
            m_options->Set(option, value);
        }

        /// <summary>
        /// Sets the property integer value for the specified key value.
        /// </summary>
        /// <param name="option">The string option</param>
        /// <param name="value">The new integer value</param>
        void Set(const std::string& option, int value)
        {
            m_options->Set(option, value);
        }

        /// <summary>
        /// Sets the property integer value for the specified key value.
        /// </summary>
        /// <param name="option">The string option</param>
        /// <param name="value">The new integer value</param>
        void Set(const std::wstring& option, int value)
        {
            m_options->Set(option, value);
        }

        // TODO: TFS#3664451 - Vision: Other VisionSessionOptions AdvancedOptions strongly-typed setting methods ...

    protected:
        /*! \cond PROTECTED */
        static std::shared_ptr<AdvancedOptions> Create()
        {
            auto ptr = new AdvancedOptions();
            return std::shared_ptr<AdvancedOptions>(ptr);
        }

        explicit AdvancedOptions() :
            m_options(PropertyCollection::Create())
        {
        }

        explicit operator AZAC_HANDLE()
        {
            return ProtectedAccess<PropertyCollection>::HandleFromPtr(m_options.get());
        }
        /*! \endcond */

    private:

        AZAC_DISABLE_COPY_AND_MOVE(AdvancedOptions);
    };

private:

    template<typename Target> using ProtectedAccess = Azure::AI::Core::Details::ProtectedAccess<Target>;

    std::shared_ptr<AdvancedOptions> m_options;

public:

    /// <summary>
    /// Initializes a new instance of the VisionSessionOptions class.
    /// </summary>
    /// <returns>The newly created VisionSessionOptions wrapped inside a std::shared_ptr</returns>
    static std::shared_ptr<VisionSessionOptions> Create()
    {
        auto ptr = new VisionSessionOptions();
        return std::shared_ptr<VisionSessionOptions>(ptr);
    }

    /// <summary>
    /// Destructs an instance of the VisionSessionOptions class.
    /// </summary>
    ~VisionSessionOptions() = default;

    /// <summary>
    /// Advanced options and parameters
    /// </summary>
    AdvancedOptions& Advanced;

    // TODO: TFS#3664462 - Vision: What VisionSessionOptions strongly typed options go here?

protected:
    /*! \cond PROTECTED */
    explicit VisionSessionOptions() :
        m_options(ProtectedAccess<AdvancedOptions>::Create()),
        Advanced(*m_options.get())
    {
    }

    explicit operator AZAC_HANDLE()
    {
        return ProtectedAccess<AdvancedOptions>::HandleFromPtr(m_options.get());
    }
    /*! \endcond */

private:

    AZAC_DISABLE_COPY_AND_MOVE(VisionSessionOptions);
};

} } } } } // Azure::AI::Vision::Core::Session
