//
// Copyright (c) Microsoft. All rights reserved.
// See https://aka.ms/azai/vision/license for the full license information.
//

#pragma once
#include <azac_api_cxx_details_property_collection.h>
#include <vision_api_c_source.h>
#include <vision_api_cxx_common.h>
#include <vision_api_cxx_frame_source.h>
#include <vision_api_cxx_frameset_source.h>
#include <vision_api_cxx_source_properties.h>
#include <vision_api_cxx_source_property.h>

namespace Azure {
namespace AI {
namespace Vision {
namespace Input {

/// <summary>
/// Represents a source of vision data, used as input to a Computer Vision operation
/// </summary>
class VisionSource
{
private:

    template<typename Target> using ProtectedAccess = Azure::AI::Core::Details::ProtectedAccess<Target>;
    AI::Core::Details::PropertyCollection<VisionSourceProperty, VisionSourceProperties> m_properties;

public:

    /// <summary>
    /// Initializes a new instance of the VisionSource class using the default capture device, such as a camera
    /// </summary>
    /// <returns>The newly created VisionSource wrapped inside a std::shared_ptr</returns>
    /// <remarks>
    /// Not yet supported for use with ImageAnalyzer
    /// </remarks>
    static std::shared_ptr<VisionSource> FromDefaultCaptureDevice()
    {
        return FromCaptureDevice("");
    }

    /// <summary>
    /// Initializes a new instance of the VisionSource class from a capture device specified by its attributes
    /// </summary>
    /// <param name="deviceAttributes">Capture device attributes (e.g. "front=true;"), or an 
    /// empty string for default capture device<</param>
    /// <returns>The newly created VisionSource wrapped inside a std::shared_ptr</returns>
    /// <remarks>
    /// Not yet supported for use with ImageAnalyzer.
    /// </remarks>
    static std::shared_ptr<VisionSource> FromCaptureDevice(const std::string& deviceAttributes)
    {
        AZAC_HANDLE handle = AZAC_HANDLE_INVALID;
        AZAC_THROW_ON_FAIL(::vision_source_handle_create(&handle, "source.device.attributes", deviceAttributes.c_str(), nullptr, nullptr));
        return FromHandle(handle);
    }

    /// <summary>
    /// Initializes a new instance of the VisionSource class from a capture device specified by its attributes
    /// </summary>
    /// <param name="deviceAttributes">Capture device attributes (e.g. "front=true;"), or an 
    /// empty string for default capture device<</param>
    /// <returns>The newly created VisionSource wrapped inside a std::shared_ptr</returns>
    /// <remarks>
    /// Not yet supported for use with ImageAnalyzer.
    /// </remarks>
    static std::shared_ptr<VisionSource> FromCaptureDevice(const std::wstring& deviceAttributes)
    {
        return FromCaptureDevice(AI::Core::Details::to_string<std::string>(deviceAttributes));
    }

    /// <summary>
    /// Initializes a new instance of the VisionSource class using a local input image or video file
    /// </summary>
    /// <param name="fileName">The full path name of the local file</param>
    /// <returns>The newly created VisionSource wrapped inside a std::shared_ptr</returns>
    static std::shared_ptr<VisionSource> FromFile(const std::string& fileName)
    {
        AZAC_HANDLE handle = AZAC_HANDLE_INVALID;
        AZAC_THROW_ON_FAIL(::vision_source_handle_create(&handle, "source.file.name", fileName.c_str(), nullptr, nullptr));
        return FromHandle(handle);
    }

    /// <summary>
    /// Initializes a new instance of the VisionSource class using a local input image or video file
    /// </summary>
    /// <param name="fileName">The full path name of the local file</param>
    /// <returns>The newly created VisionSource wrapped inside a std::shared_ptr</returns>
    static std::shared_ptr<VisionSource> FromFile(const std::wstring& fileName)
    {
        return FromFile(AI::Core::Details::to_string<std::string>(fileName));
    }

    /// <summary>
    /// Initializes a new instance of the VisionSource class using the specified internet-accessible URL
    /// </summary>
    /// <param name="url">An accessible URL for the input media</param>
    /// <returns>The newly created VisionSource wrapped inside a std::shared_ptr</returns>
    static std::shared_ptr<VisionSource> FromUrl(const std::string& url)
    {
        AZAC_HANDLE handle = AZAC_HANDLE_INVALID;
        AZAC_THROW_ON_FAIL(::vision_source_handle_create(&handle, "source.url.name", url.c_str(), nullptr, nullptr));
        return FromHandle(handle);
    }

    /// <summary>
    /// Initializes a new instance of the VisionSource class using the specified internet-accessible URL
    /// </summary>
    /// <param name="url">An accessible URL for the input media</param>
    /// <returns>The newly created VisionSource wrapped inside a std::shared_ptr</returns>
    static std::shared_ptr<VisionSource> FromUrl(const std::wstring& url)
    {
        return FromUrl(AI::Core::Details::to_string<std::string>(url));
    }

    /// <summary>
    /// Initializes a new instance of the VisionSource class with a FrameSource
    /// </summary>
    /// <param name="frameSource">A FrameSource obtained via FrameSource::FromFormat()</param>
    /// <returns>The newly created VisionSource wrapped inside a std::shared_ptr</returns>
    /// <remarks>
    /// Not yet supported for use with ImageAnalyzer
    /// </remarks>
    static std::shared_ptr<VisionSource> FromFrameSource(const std::shared_ptr<Frames::FrameSource>& frameSource)
    {
        AZAC_HANDLE handle = AZAC_HANDLE_INVALID;
        AZAC_HANDLE frameSourceHandle = ProtectedAccess<Frames::FrameSource>::HandleFromPtr(frameSource.get());
        AZAC_THROW_ON_FAIL(::vision_source_handle_create(&handle, nullptr, nullptr, frameSourceHandle, nullptr));
        return FromHandle(handle);
    }

    /// <summary>
    /// Initializes a new instance of the VisionSource class with a FrameSource
    /// </summary>
    /// <param name="frameSetSource">A FrameSetSource obtained via FrameSetSource::FromFormat()</param>
    /// <returns>The newly created VisionSource wrapped inside a std::shared_ptr</returns>
    /// <remarks>
    /// Not yet supported for use with ImageAnalyzer
    /// </remarks>
    static std::shared_ptr<VisionSource> FromFrameSetSource(const std::shared_ptr<Frames::FrameSetSource>& frameSetSource)
    {
        AZAC_HANDLE handle = AZAC_HANDLE_INVALID;
        AZAC_HANDLE frameSourceHandle = ProtectedAccess<Frames::FrameSetSource>::HandleFromPtr(frameSetSource.get());
        AZAC_THROW_ON_FAIL(::vision_source_handle_create(&handle, nullptr, nullptr, frameSourceHandle, nullptr));
        return FromHandle(handle);
    }

    /// <summary>
    /// Destructs an instance of the VisionSource class.
    /// </summary>
    virtual ~VisionSource()
    {
        if (vision_source_handle_is_valid(m_source))
        {
            ::vision_source_handle_release(m_source);
            m_source = AZAC_HANDLE_INVALID;
        }
    };

    /// <summary>
    /// Gets a collection of additional VisionSource properties.
    /// </summary>
    Input::VisionSourceProperties& Properties;

protected:
    /*! \cond PROTECTED */
    static std::shared_ptr<VisionSource> FromHandle(AZAC_HANDLE handle)
    {
        auto ptr = new VisionSource(handle);
        return std::shared_ptr<VisionSource>(ptr);
    }

    explicit VisionSource(AZAC_HANDLE source) :
        m_properties(source, [](auto handle, auto* properties) { return vision_source_properties_handle_get(handle, properties); }),
        Properties(m_properties),
        m_source(source)
    {
    }

    explicit operator AZAC_HANDLE() { return m_source; }
    /*! \endcond */

private:

    AZAC_DISABLE_DEFAULT_CTORS(VisionSource);

    AZAC_HANDLE m_source;
};

} } } } // Azure::AI::Vision::Input
